import java.io.*;
import java.net.*;



public class MissingClassTest {

    /**
     * Entry point to this simple test.
     */
    public static void main(String args[]) throws ClassNotFoundException {
        System.out.println("MissingClassTest.java");
        new MissingClass();
        System.exit(0);
    }
}

// finito


